#include "console.h"
#include "color.h"

#include <string.h>

static struct console_data_s *default_console;

void (*my_puts)(const char *s);
void (*my_putstring)(const char *s);
void (*my_putchar)(char c);

void c_putstring(const char *s)
{
	console_puts(default_console, s);
}

void c_puts(const char *s)
{
	console_puts(default_console, s);
	console_putc(default_console, '\n');
}

void c_putchar(char c)
{
	console_putc(default_console, c);
}

#if 0 
void s_putc(unsigned char c);

void s_putstring(const char *s)
{
	while (*s)
	{
		if (*s == '\n')
			s_putc('\r');
		s_putc(*s++);
	}
}

void s_puts(const char *s)
{
	s_putstring(s);
	s_putstring("\n");
}

void s_putchar(char c)
{
	s_putc(c);
}
#endif

void puts(const char *s)
{
	if (my_puts)
		my_puts(s);
}

void putstring(const char *s)
{
	if (my_putstring)
		my_putstring(s);
}

void putchar(char c)
{
	if (my_putchar)
		my_putchar(c);
}

int set_color(int background, int foreground)
{
	default_console->foreground = foreground;
	default_console->background = background;
	return 0;
}

struct console_data_s console;
extern unsigned char console_font_8x8[];
extern unsigned char console_font_8x16[];

void console_init(void *framebuffer, int xres, int yres, int stride)
{
	struct console_data_s *con = &console;
	con->framebuffer = framebuffer;
	con->xres = xres;
	con->yres = yres;
	con->border_left = 0;
	con->border_top  = 0;
	con->border_right = con->xres;
	con->border_bottom = con->yres;
	con->stride = stride;
	con->cursor_x = con->cursor_y = 0;
	
	con->font = console_font_8x16;
	
	con->foreground = COLOR_WHITE;
	con->background = COLOR_BLACK;
	
	con->scrolled_lines = 0;
	int c = con->xres * con->yres / 2;
	unsigned long *p = (unsigned long*)con->framebuffer;
	while (c--)
		*p++ = con->background;
	default_console = con;
	
	my_puts = c_puts;
	my_putstring = c_putstring;
	my_putchar = c_putchar;
	
/*	my_puts = s_puts;
	my_putstring = s_putstring;
	my_putchar = s_putchar; */
}

#define FONT_XSIZE  8
#define FONT_YSIZE  16
#define FONT_XFACTOR 1
#define FONT_YFACTOR 1
#define FONT_XGAP   2
#define FONT_YGAP   0

void console_drawc(struct console_data_s *con, int x, int y, unsigned char c)
{
	x >>= 1;
	int ax, ay;
	unsigned long *ptr = (unsigned long*)(con->framebuffer + con->stride * y + x * 4);
	for (ay = 0; ay < FONT_YSIZE; ay++)
#if FONT_XFACTOR == 2
		for (ax = 0; ax < 8; ax++)
		{
			unsigned long color;
			if ((con->font[c * FONT_YSIZE + ay] << ax) & 0x80)
				color = con->foreground;
			else
				color = con->background;
#if FONT_YFACTOR == 2
				// pixel doubling: we write u32
			ptr[ay * 2 * con->stride/4 + ax] = color;
				// line doubling
			ptr[(ay * 2 +1) * con->stride/4 + ax] = color;
#else
			ptr[ay * con->stride/4 + ax] = color;
#endif
		}
#else
		for (ax = 0; ax < 4; ax ++)
		{
			unsigned long color[2];
			int bits = (con->font[c * FONT_YSIZE + ay] << (ax*2));
			if (bits & 0x80)
				color[0] = con->foreground;
			else
				color[0] = con->background;
			if (bits & 0x40)
				color[1] = con->foreground;
			else
				color[1] = con->background;
			ptr[ay * con->stride/4 + ax] = (color[0] & 0xFFFF00FF) | (color[1] & 0x0000FF00);
		}
#endif
}

void console_puts(struct console_data_s *con, const char *string)
{
	while (*string)
		console_putc(con, *string++);
}

void console_putc(struct console_data_s *con, char c)
{
	switch (c)
	{
	case '\n':
		con->cursor_y += FONT_YSIZE * FONT_YFACTOR + FONT_YGAP;
		con->cursor_x = con->border_left;
		break;
	case '\r':
		con->cursor_x = con->border_left;
		break;
	default:
		console_drawc(con, con->cursor_x, con->cursor_y, c);
		con->cursor_x += FONT_XSIZE * FONT_XFACTOR + FONT_XGAP;
		if ((con->cursor_x + (FONT_XSIZE * FONT_XFACTOR)) > con->border_right)
		{
			con->cursor_y += FONT_YSIZE * FONT_YFACTOR + FONT_YGAP;
			con->cursor_x = con->border_left;
		}
	}
	if ((con->cursor_y + FONT_YSIZE * FONT_YFACTOR) >= con->border_bottom)
	{
#if 0
		if (con->scrolled_lines == con->yres / (FONT_YSIZE * FONT_YFACTOR + FONT_YGAP))
		{
			struct PAD pad;
			while (1)
			{
				PAD_ReadState(&pad, 0);
				if (pad.Digital.A)
					break;
			}
			while (1)
			{
				PAD_ReadState(&pad, 0);
				if (!pad.Digital.A)
					break;
			}
			con->scrolled_lines = 0;
		}
		con->scrolled_lines++;
#endif
		memcpy(con->framebuffer, 
				con->framebuffer + con->stride * (FONT_YSIZE * FONT_YFACTOR + FONT_YGAP), 
				con->stride * con->yres - FONT_YSIZE);
		int cnt = (con->stride * (FONT_YSIZE * FONT_YFACTOR + FONT_YGAP)) / 4;
		int a = 0;
		unsigned long *ptr = con->framebuffer + con->stride * (con->yres - FONT_YSIZE);
		while (cnt--)
			 *ptr++ = con->background;
		con->cursor_y -= FONT_YSIZE * FONT_YFACTOR + FONT_YGAP;
	}
}
